from datetime import datetime, timedelta

import pytz
from asgiref.sync import sync_to_async

from consoleme.config import config
from consoleme.handlers.base import BaseHandler
from consoleme.lib.crypto import Crypto
from consoleme.lib.jwt import generate_jwt_token
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.lib.saml import init_saml_auth, prepare_tornado_request_for_saml

if config.get("auth.get_user_by_saml"):
    from onelogin.saml2.utils import OneLogin_Saml2_Utils

stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()
log = config.get_logger()
crypto = Crypto()
auth = get_plugin_by_name(config.get("plugins.auth", "default_auth"))()


class SamlHandler(BaseHandler):
    def check_xsrf_cookie(self):
        pass

    async def post(self, endpoint):
        req = await prepare_tornado_request_for_saml(self.request)
        auth = await init_saml_auth(req)

        if "sso" in endpoint:
            return self.redirect(auth.login())
        elif "acs" in endpoint:
            auth.process_response()
            errors = auth.get_errors()
            not_auth_warn = not await sync_to_async(auth.is_authenticated)()
            if not_auth_warn:
                self.write("User is not authenticated")
                await self.finish()
                return
            if len(errors) == 0:

                saml_attributes = await sync_to_async(auth.get_attributes)()
                email = saml_attributes[
                    config.get("get_user_by_saml_settings.attributes.email")
                ]
                if isinstance(email, list) and len(email) > 0:
                    email = email[0]
                groups = saml_attributes.get(
                    config.get("get_user_by_saml_settings.attributes.groups"), []
                )

                self_url = await sync_to_async(OneLogin_Saml2_Utils.get_self_url)(req)
                if config.get("auth.set_auth_cookie"):
                    expiration = datetime.utcnow().replace(tzinfo=pytz.UTC) + timedelta(
                        minutes=config.get("jwt.expiration_minutes", 60)
                    )
                    encoded_cookie = await generate_jwt_token(
                        email, groups, exp=expiration
                    )
                    self.set_cookie(
                        config.get("auth_cookie_name", "consoleme_auth"),
                        encoded_cookie,
                        expires=expiration,
                        secure=config.get(
                            "auth.cookie.secure",
                            "https://" in config.get("url"),
                        ),
                        httponly=config.get("auth.cookie.httponly", True),
                        samesite=config.get("auth.cookie.samesite", True),
                    )
                if (
                    "RelayState" in self.request.arguments
                    and self_url
                    != self.request.arguments["RelayState"][0].decode("utf-8")
                ):
                    return self.redirect(
                        auth.redirect_to(
                            self.request.arguments["RelayState"][0].decode("utf-8")
                        )
                    )



import base64
import sys
from datetime import datetime, timedelta

import jwt
import pytz
import tornado.httpclient
import ujson as json
from jwt.algorithms import ECAlgorithm, RSAAlgorithm
from jwt.exceptions import DecodeError
from tornado import httputil

from consoleme.config import config
from consoleme.exceptions.exceptions import (
    MissingConfigurationValue,
    UnableToAuthenticate,
)
from consoleme.lib.generic import should_force_redirect
from consoleme.lib.jwt import generate_jwt_token

log = config.get_logger()


async def populate_oidc_config():
    http_client = tornado.httpclient.AsyncHTTPClient()
    metadata_url = config.get("get_user_by_oidc_settings.metadata_url")

    if metadata_url:
        res = await http_client.fetch(
            metadata_url,
            method="GET",
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "Accept": "application/json",
            },
        )
        oidc_config = json.loads(res.body)
    else:
        authorization_endpoint = config.get(
            "get_user_by_oidc_settings.authorization_endpoint"
        )
        token_endpoint = config.get("get_user_by_oidc_settings.token_endpoint")
        jwks_uri = config.get("get_user_by_oidc_settings.jwks_uri")
        if not (authorization_endpoint or token_endpoint or jwks_uri):
            raise MissingConfigurationValue("Missing OIDC Configuration.")
        oidc_config = {
            "authorization_endpoint": authorization_endpoint,
            "token_endpoint": token_endpoint,
            "jwks_uri": jwks_uri,
        }
    client_id = config.get("oidc_secrets.client_id")
    client_secret = config.get("oidc_secrets.secret")
    if not (client_id or client_secret):
        raise MissingConfigurationValue("Missing OIDC Secrets")
    oidc_config["client_id"] = client_id
    oidc_config["client_secret"] = client_secret
    oidc_config["jwt_keys"] = {}
    jwks_uris = [oidc_config["jwks_uri"]]
    jwks_uris.extend(config.get("get_user_by_oidc_settings.extra_jwks_uri", []))
    for jwks_uri in jwks_uris:
        # Fetch jwks_uri for jwt validation
        res = await http_client.fetch(
            jwks_uri,
            method="GET",
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "Accept": "application/json",
            },
        )
        jwks_data = json.loads(res.body)
        for k in jwks_data["keys"]:
            key_type = k["kty"]
            key_id = k["kid"]
            if key_type == "RSA":
                oidc_config["jwt_keys"][key_id] = RSAAlgorithm.from_jwk(json.dumps(k))
            elif key_type == "EC":
                oidc_config["jwt_keys"][key_id] = ECAlgorithm.from_jwk(json.dumps(k))
            else:
                raise MissingConfigurationValue(
                    f"OIDC/OAuth2 key type not recognized. Detected key type: {key_type}."
                )
    return oidc_config


async def authenticate_user_by_oidc(request):
    email = None
    groups = []
    decoded_access_token = {}
    oidc_config = await populate_oidc_config()
    function = f"{__name__}.{sys._getframe().f_code.co_name}"
    log_data = {"function": function}
    code = request.get_argument("code", None)
    protocol = request.request.protocol
    if "https://" in config.get("url"):
        # If we're behind a load balancer that terminates tls for us, request.request.protocol will be "http://" and our
        # oidc redirect will be invalid
        protocol = "https"
    force_redirect = await should_force_redirect(request.request)

    # The endpoint where we want our OIDC provider to redirect us back to perform auth
    oidc_redirect_uri = f"{protocol}://{request.request.host}/auth"

    # The endpoint where the user wants to be sent after authentication. This will be stored in the state
    after_redirect_uri = request.request.arguments.get("redirect_url", [""])[0]
    if not after_redirect_uri:
        after_redirect_uri = request.request.arguments.get("state", [""])[0]
    if after_redirect_uri and isinstance(after_redirect_uri, bytes):
        after_redirect_uri = after_redirect_uri.decode("utf-8")
    if not after_redirect_uri:
        after_redirect_uri = config.get("url", f"{protocol}://{request.request.host}/")

    if not code:
        args = {"response_type": "code"}
        client_scope = config.get("oidc_secrets.client_scope")
        if request.request.uri is not None:
            args["redirect_uri"] = oidc_redirect_uri
        args["client_id"] = oidc_config["client_id"]
        if client_scope:
            args["scope"] = " ".join(client_scope)
        # TODO: Sign and verify redirect URI with expiration
        args["state"] = after_redirect_uri

        if force_redirect:
            request.redirect(
                httputil.url_concat(oidc_config["authorization_endpoint"], args)
            )
        else:
            request.set_status(403)
            request.write(
                {
                    "type": "redirect",
                    "redirect_url": httputil.url_concat(
                        oidc_config["authorization_endpoint"], args
                    ),
                    "reason": "unauthenticated",
                    "message": "User is not authenticated. Redirect to authenticate",
                }
            )
        request.finish()
        return
    try:
        # exchange the authorization code with the access token
        http_client = tornado.httpclient.AsyncHTTPClient()
        grant_type = config.get(
            "get_user_by_oidc_settings.grant_type", "authorization_code"
        )
        authorization_header = (
            f"{oidc_config['client_id']}:{oidc_config['client_secret']}"
        )
        authorization_header_encoded = base64.b64encode(
            authorization_header.encode("UTF-8")
        ).decode("UTF-8")
        url = f"{oidc_config['token_endpoint']}"
        client_scope = config.get("oidc_secrets.client_scope")
        if client_scope:
            client_scope = " ".join(client_scope)
        token_exchange_response = await http_client.fetch(
            url,
            method="POST",
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "Authorization": "Basic %s" % authorization_header_encoded,
                "Accept": "application/json",
            },
            body=f"grant_type={grant_type}&code={code}&redirect_uri={oidc_redirect_uri}&scope={client_scope}",
        )

        token_exchange_response_body_dict = json.loads(token_exchange_response.body)

        id_token = token_exchange_response_body_dict.get(
            config.get("get_user_by_oidc_settings.id_token_response_key", "id_token")
        )
        access_token = token_exchange_response_body_dict.get(
            config.get(
                "get_user_by_oidc_settings.access_token_response_key", "access_token"
            )
        )

        jwt_verify = config.get("get_user_by_oidc_settings.jwt_verify", True)
        if jwt_verify:
            header = jwt.get_unverified_header(id_token)
            key_id = header["kid"]
            algorithm = header["alg"]
            if algorithm == "none" or not algorithm:
                raise UnableToAuthenticate(
                    "ID Token header does not specify a signing algorithm."
                )
            pub_key = oidc_config["jwt_keys"][key_id]
            # This will raises errors if the audience isn't right or if the token is expired or has other errors.
            decoded_id_token = jwt.decode(
                id_token,
                pub_key,
                audience=oidc_config["client_id"],
                algorithms=algorithm,
            )

            email = decoded_id_token.get(
                config.get("get_user_by_oidc_settings.jwt_email_key", "email")
            )

            # For google auth, the access_token does not contain JWT-parsable claims.
            if config.get(
                "get_user_by_oidc_settings.get_groups_from_access_token", True
            ):
                try:
                    header = jwt.get_unverified_header(access_token)
                    key_id = header["kid"]
                    algorithm = header["alg"]
                    if algorithm == "none" or not algorithm:
                        raise UnableToAuthenticate(
                            "Access Token header does not specify a signing algorithm."
                        )
                    pub_key = oidc_config["jwt_keys"][key_id]
                    # This will raises errors if the audience isn't right or if the token is expired or has other
                    # errors.
                    decoded_access_token = jwt.decode(
                        access_token,
                        pub_key,
                        audience=config.get(
                            "get_user_by_oidc_settings.access_token_audience"
                        ),
                        algorithms=algorithm,
                    )
                except (DecodeError, KeyError) as e:
                    # This exception occurs when the access token is opaque or otherwise not JWT-parsable.
                    # It is expected with some IdPs.
                    log.debug(
                        {
                            **log_data,
                            "message": (
                                "Unable to derive user's groups from access_token. Attempting to get groups through "
                                "userinfo endpoint. "
                            ),
                            "error": e,
                            "user": email,
                        }
                    )
                    log.debug(log_data, exc_info=True)
                    groups = []
        else:
            decoded_id_token = jwt.decode(id_token, verify=jwt_verify)
            decoded_access_token = jwt.decode(access_token, verify=jwt_verify)

        email = email or decoded_id_token.get(
            config.get("get_user_by_oidc_settings.jwt_email_key", "email")
        )

        if not email:
            raise UnableToAuthenticate("Unable to determine user from ID Token")

        groups = (
            groups
            or decoded_access_token.get(
                config.get("get_user_by_oidc_settings.jwt_groups_key", "groups"),
            )
            or decoded_id_token.get(
                config.get("get_user_by_oidc_settings.jwt_groups_key", "groups"), []
            )
        )

        if (
            not groups
            and oidc_config.get("userinfo_endpoint")
            and config.get(
                "get_user_by_oidc_settings.get_groups_from_userinfo_endpoint", True
            )
        ):
            user_info = await http_client.fetch(
                oidc_config["userinfo_endpoint"],
                method="GET",
                headers={
                    "Content-Type": "application/x-www-form-urlencoded",
                    "Authorization": "Bearer %s" % access_token,
                    "Accept": "application/json",
                },
            )
            groups = json.loads(user_info.body).get(
                config.get("get_user_by_oidc_settings.user_info_groups_key", "groups"),
                [],
            )

        if config.get("auth.set_auth_cookie"):
            expiration = datetime.utcnow().replace(tzinfo=pytz.UTC) + timedelta(
                minutes=config.get("jwt.expiration_minutes", 60)
            )
            encoded_cookie = await generate_jwt_token(email, groups, exp=expiration)
            request.set_cookie(
                config.get("auth_cookie_name", "consoleme_auth"),
                encoded_cookie,
                expires=expiration,
                secure=config.get(
                    "auth.cookie.secure",
                    "https://" in config.get("url"),
                ),
                httponly=config.get("auth.cookie.httponly", True),
                samesite=config.get("auth.cookie.samesite", True),
            )

        request.redirect(after_redirect_uri)
        request.finish()

    except Exception as e:
        log_data["error"] = e
        log.error(log_data, exc_info=True)
        return



import sys
from typing import List

import ujson as json
from tornado.httpclient import AsyncHTTPClient, HTTPClientError

from consoleme.config import config
from consoleme.exceptions.exceptions import MissingConfigurationValue
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.models import CloudAccountModel, CloudAccountModelArray

log = config.get_logger()
stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()


async def retrieve_accounts_from_swag() -> CloudAccountModelArray:
    function: str = f"{sys._getframe().f_code.co_name}"
    expected_owners: List = config.get(
        "retrieve_accounts_from_swag.expected_owners", []
    )

    swag_base_url = config.get("retrieve_accounts_from_swag.base_url")
    if not swag_base_url:
        raise MissingConfigurationValue("Unable to find Swag URL in configuration")
    swag_url = swag_base_url + "api/1/accounts"

    try:
        http_client = AsyncHTTPClient(force_instance=True)
        resp = await http_client.fetch(
            swag_url,
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        )

    except (ConnectionError, HTTPClientError) as e:
        log.error(
            {
                "message": "Unable to connect to SWAG",
                "error": str(e),
                "function": function,
            },
            exc_info=True,
        )
        stats.count(f"{function}.connectionerror")
        raise
    swag_accounts = json.loads(resp.body)
    cloud_accounts = []
    for account in swag_accounts:
        # Ignore third party accounts
        if expected_owners and account.get("owner") not in expected_owners:
            continue
        account_status = account["account_status"]
        sync_enabled = False
        if account_status == "ready":
            account_status = "active"
            sync_enabled = True
        cloud_accounts.append(
            CloudAccountModel(
                id=account["id"],
                name=account["name"],
                email=account["email"],
                status=account_status,
                sync_enabled=sync_enabled,
                sensitive=account["sensitive"],
                environment=account["environment"],
                aliases=account["aliases"],
                type="aws",
            )
        )
    return CloudAccountModelArray(accounts=cloud_accounts)



"""Docstring in public module."""
import copy
import json
import os
import sys
from datetime import datetime, timedelta
from unittest import TestCase

APP_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
sys.path.append(os.path.join(APP_ROOT, ".."))


class TestCelerySync(TestCase):
    def setUp(self):
        from consoleme.celery_tasks import celery_tasks as celery

        self.celery = celery

    def cache_iam_resources_for_account(self):
        from consoleme.config.config import CONFIG
        from consoleme.lib.dynamo import IAMRoleDynamoHandler
        from consoleme.lib.redis import RedisHandler

        red = RedisHandler().redis_sync()

        # Set the config value for the redis cache location
        old_config = copy.deepcopy(CONFIG.config)
        CONFIG.config = {
            **CONFIG.config,
            "aws": {
                **CONFIG.config.get("aws", {}),
                "iamroles_redis_key": "cache_iam_resources_for_account",
            },
            "cache_iam_resources_across_accounts": {
                "all_roles_combined": {
                    "s3": {
                        "file": "cache_iam_resources_for_account.json.gz",
                    }
                }
            },
        }

        # Clear out the existing cache from Redis:
        red.delete("cache_iam_resources_for_account")
        # Run it:
        self.celery.cache_iam_resources_for_account("123456789012")

        # Verify that everything is there:
        dynamo = IAMRoleDynamoHandler()

        results = dynamo.role_table.scan(TableName="consoleme_iamroles_global")

        remaining_roles = [
            "arn:aws:iam::123456789012:role/ConsoleMe",
            "arn:aws:iam::123456789012:role/cm_someuser_N",
            "arn:aws:iam::123456789012:role/awsaccount_user",
            "arn:aws:iam::123456789012:role/TestInstanceProfile",
            "arn:aws:iam::123456789012:role/rolename",
        ] + [f"arn:aws:iam::123456789012:role/RoleNumber{num}" for num in range(0, 10)]

        self.assertEqual(results["Count"], len(remaining_roles))
        self.assertEqual(results["Count"], red.hlen("cache_iam_resources_for_account"))

        for i in results["Items"]:
            remaining_roles.remove(i["arn"])
            self.assertEqual(i["accountId"], "123456789012")
            self.assertGreater(int(i["ttl"]), 0)
            self.assertIsNotNone(json.loads(i["policy"]))
            self.assertEqual(
                json.loads(red.hget("cache_iam_resources_for_account", i["arn"]))[
                    "policy"
                ],
                i["policy"],
            )

        # Should all be accounted for:
        self.assertEqual(remaining_roles, [])

        # We should have the same data in redis on all regions, this time coming from DDB
        old_conf_region = self.celery.config.region
        self.celery.config.region = "us-east-1"

        # Clear out the existing cache from Redis:
        red.delete("cache_iam_resources_for_account")

        # This should spin off extra fake celery tasks
        res = self.celery.cache_iam_resources_across_accounts()
        self.assertEqual(
            res,
            {
                "function": "consoleme.celery_tasks.celery_tasks.cache_iam_resources_across_accounts",
                "cache_key": "cache_iam_resources_for_account",
                "num_roles": 15,
                "num_accounts": 1,
            },
        )  # This should spin off extra fake celery tasks
        res = self.celery.cache_iam_resources_across_accounts()
        self.assertEqual(
            res,
            {
                "function": "consoleme.celery_tasks.celery_tasks.cache_iam_resources_across_accounts",
                "cache_key": "cache_iam_resources_for_account",
                "num_roles": 15,
                "num_accounts": 1,
            },
        )

        # Reset the config value:
        self.celery.config.region = old_conf_region
        CONFIG.config = old_config

    def test_clear_old_redis_iam_cache(self):
        from consoleme.config.config import CONFIG
        from consoleme.lib.redis import RedisHandler

        red = RedisHandler().redis_sync()

        self.celery.REDIS_IAM_COUNT = 3

        # Clear out the existing cache from Redis:
        red.delete("cache_iam_resources_for_account_expiration")

        # Set the config value for the redis cache location
        old_value = CONFIG.config["aws"].pop("iamroles_redis_key", None)
        CONFIG.config["aws"][
            "iamroles_redis_key"
        ] = "cache_iam_resources_for_account_expiration"

        # Add in some dummy IAM roles with a TTL that is more than 6 hours old:
        old_ttl = int((datetime.utcnow() - timedelta(hours=6, seconds=5)).timestamp())

        # 13 items / 3 = 5 iterations -- all of these roles should be cleaned up:
        for i in range(0, 13):
            role_entry = {
                "arn": f"arn:aws:iam::123456789012:role/RoleNumber{i}",
                "name": f"RoleNumber{i}",
                "accountId": "123456789012",
                "ttl": old_ttl,
                "policy": "{}",
            }
            self.celery._add_role_to_redis(
                "cache_iam_resources_for_account_expiration", role_entry
            )

        # Add a role with a current TTL -- this should not be cleaned up:
        role_entry = {
            "arn": "arn:aws:iam::123456789012:role/RoleNumber99",
            "name": "RoleNumber99",
            "accountId": "123456789012",
            "ttl": int(datetime.utcnow().timestamp()),
            "policy": "{}",
        }
        self.celery._add_role_to_redis(
            "cache_iam_resources_for_account_expiration", role_entry
        )

        # Nothing should happen if we are not in us-west-2:
        old_conf_region = self.celery.config.region
        self.celery.config.region = "us-east-1"

        self.celery.clear_old_redis_iam_cache()
        self.assertEqual(red.hlen("cache_iam_resources_for_account_expiration"), 14)

        # With the proper region:
        self.celery.config.region = "us-west-2"
        self.celery.clear_old_redis_iam_cache()

        # Verify:
        self.assertEqual(red.hlen("cache_iam_resources_for_account_expiration"), 1)
        self.assertIsNotNone(
            red.hget(
                "cache_iam_resources_for_account_expiration",
                "arn:aws:iam::123456789012:role/RoleNumber99",
            )
        )

        # Clear out the existing cache from Redis:
        red.delete("cache_iam_resources_for_account_expiration")

        # Reset the config values:
        self.celery.config.region = old_conf_region
        self.celery.REDIS_IAM_COUNT = 1000
        if not old_value:
            del CONFIG.config["aws"]["iamroles_redis_key"]
        else:
            CONFIG.config["aws"]["iamroles_redis_key"] = old_value

    def test_trigger_credential_mapping_refresh_from_role_changes(self):
        res = self.celery.trigger_credential_mapping_refresh_from_role_changes()
        self.assertEqual(
            res,
            {
                "function": "consoleme.celery_tasks.celery_tasks.trigger_credential_mapping_refresh_from_role_changes",
                "message": "Successfully checked role changes",
                "num_roles_changed": 2,
            },
        )

    def test_cache_cloudtrail_denies(self):
        res = self.celery.cache_cloudtrail_denies()
        self.assertEqual(
            res,
            {
                "function": "consoleme.celery_tasks.celery_tasks.cache_cloudtrail_denies",
                "message": "Successfully cached cloudtrail denies",
                "num_new_cloudtrail_denies": 1,
                "num_cloudtrail_denies": 1,
            },
        )



import sentry_sdk
import tornado.escape
import tornado.web
import ujson as json
from asgiref.sync import sync_to_async
from marshmallow import Schema, ValidationError, fields, validates_schema

from consoleme.config import config
from consoleme.exceptions.exceptions import CertTooOldException
from consoleme.handlers.base import BaseMtlsHandler
from consoleme.lib.account_indexers import get_cloud_account_model_array
from consoleme.lib.crypto import Crypto
from consoleme.lib.duo import duo_mfa_user
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.models import Environment

stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()
log = config.get_logger()
crypto = Crypto()
aws = get_plugin_by_name(config.get("plugins.aws", "default_aws"))()
internal_config = config.config_plugin
auth = get_plugin_by_name(config.get("plugins.auth", "default_auth"))()
group_mapping = get_plugin_by_name(
    config.get("plugins.group_mapping", "default_group_mapping")
)()
internal_policies = get_plugin_by_name(
    config.get("plugins.internal_policies", "default_policies")
)()


class CredentialsSchema(Schema):
    requested_role = fields.Str()
    user_role = fields.Boolean(default=False, missing=False)
    app_name = fields.Str()
    account = fields.Str()
    console_only = fields.Boolean(default=False, missing=False)
    no_ip_restrictions = fields.Boolean(default=False, missing=False)
    custom_ip_restrictions = fields.List(fields.Str(), required=False)

    @validates_schema
    def validate_minimum(self, data, *args, **kwargs):
        """Validate that the minimum fields are supplied."""
        if (
            not data.get("requested_role")
            and not data.get("user_role")
            and not data.get("app_name")
        ):
            raise ValidationError(
                "Must supply either a requested_role, or a user_role/account combo, or an app_name."
            )

        return data

    @validates_schema
    def validate_dynamic_role_request(self, data, *args, **kwargs):
        if data.get("user_role"):
            if data.get("requested_role"):
                raise ValidationError(
                    "Cannot specify both a requested_role and user_role."
                )

            if not data.get("account"):
                raise ValidationError("Must specify an account.")

        return data

    @validates_schema
    def validate_app_name_request(self, data, *args, **kwargs):
        """Validate that if an app name is provided, then a requested role / user role aren't specified"""
        if data.get("app_name"):
            if data.get("requested_role") or data.get("user_role"):
                raise ValidationError(
                    "Cannot specify requested_role/user_role and an app name."
                )

        return data


credentials_schema = CredentialsSchema()


class GetCredentialsHandler(BaseMtlsHandler):
    """Main consoleme api handler."""

    def check_xsrf_cookie(self):
        pass

    def initialize(self):
        self.user = None
        self.eligible_roles = []

    async def raise_if_certificate_too_old(self, role, log_data=None):
        log_data = {} if not log_data else log_data
        try:
            max_cert_age = await group_mapping.get_max_cert_age_for_role(role)
            max_cert_age_message = config.get(
                "errors.custom_max_cert_age_message", "Please refresh your certificate."
            )
        except Exception as e:
            sentry_sdk.capture_exception()
            log_data["error"] = e
            log_data[
                "message"
            ] = "Failed to get max MTLS certificate age. Returning default value of 1 day"
            max_cert_age = 1  # Default to one day expiration if we fail to get max certificate age info
        max_cert_age_seconds = max_cert_age * (24 * 60 * 60)  # Seconds in a day
        try:
            if self.current_cert_age > max_cert_age_seconds:
                raise CertTooOldException(
                    f"MTLS certificate is too old. The role you selected requires a max cert "
                    f"age of {max_cert_age} days. "
                    f"{max_cert_age_message}"
                )
        except CertTooOldException as e:
            log_data["message"] = "Unable to get credentials for user"
            log_data["eligible_roles"] = self.eligible_roles
            log.warning(log_data, exc_info=True)
            stats.count(
                "GetCredentialsHandler.post.exception",
                tags={"user": self.user, "requested_role": role, "authorized": False},
            )
            error = {
                "code": "905",
                "message": (
                    f"MTLS certificate is too old. {max_cert_age_message}. "
                    f"Max cert age for {role} is {max_cert_age} days."
                ),
                "requested_role": role,
                "exception": str(e),
                "request_id": self.request_uuid,
            }
            self.set_status(403)
            self.write(error)
            await self.finish()
            raise

    async def _get_the_requested_role(self, request: dict, log_data: dict) -> str:
        """Get the requested role to complete the credentials fetching."""
        if request.get("requested_role"):
            return request["requested_role"]
        elif request.get("app_name"):
            role_models = await internal_policies.get_roles_associated_with_app(
                request["app_name"]
            )
            if not role_models:
                stats.count(
                    "GetCredentialsHandler.post",
                    tags={
                        "user": self.user,
                        "user_role": False,
                        "app_name": request["app_name"],
                    },
                )
                log_data["message"] = "No matching roles for provided app name."
                log.warning(log_data)
                error = {
                    "code": "900",
                    "message": "No matching roles for provided app name.",
                    "request_id": self.request_uuid,
                }
                self.set_status(400)
                self.write(error)
                await self.finish()
                return ""

            account_ids = set()
            # If an account was passed into this function, we can use that
            if "account" in request:
                am = await group_mapping.get_account_mappings()
                if request["account"] in am["ids_to_names"].keys():
                    # Account ID was passed in directly
                    account_ids.add(request["account"])
                else:
                    # Might be a friendly name, have to check
                    if not am["names_to_ids"].get(request["account"]):
                        stats.count(
                            "GetCredentialsHandler.post.error",
                            tags={
                                "user": self.user,
                                "user_role": False,
                                "account": request["account"],
                            },
                        )
                        log_data["message"] = "Can't find the passed in account."
                        log.warning(log_data)
                        error = {
                            "code": "906",
                            "message": "No matching account for provided account.",
                            "request_id": self.request_uuid,
                        }
                        self.set_status(400)
                        self.write(error)
                        await self.finish()
                        return ""

                    account_ids.add(am["names_to_ids"][request["account"]])

            if not account_ids:
                # no account id was passed in, check to see if we can "smartly" determine the account.
                # Preference will be given to test accounts
                filtered_accounts = await get_cloud_account_model_array(
                    environment=Environment.test.value
                )
                # convert to set for O(1) lookup
                for account in filtered_accounts.accounts:
                    account_ids.add(account.id)

            potential_arns = []
            # for all roles associated with app, find the one that is also an account in potential account ids
            for role in role_models:
                if role.account_id in account_ids:
                    potential_arns.append(role.arn)

            if len(potential_arns) != 1:
                # if length isn't exactly 1, then it's an error either way (0 or more than 1)
                if len(potential_arns) == 0:
                    code = "900"
                    message = "No matching roles"
                else:
                    code = "901"
                    message = "More than one matching role"

                stats.count(
                    "GetCredentialsHandler.post.error",
                    tags={
                        "user": self.user,
                        "user_role": False,
                    },
                )
                log_data["message"] = message
                log.warning(log_data)
                error = {
                    "code": code,
                    "message": message,
                    "request_id": self.request_uuid,
                }
                self.set_status(400)
                self.write(error)
                await self.finish()
                return ""

            # if here, then success, we found exactly 1 ARN
            return potential_arns[0]

        else:
            # Check that the account exists:
            am = await group_mapping.get_account_mappings()

            # First, check if an account ID was passed in:
            if request["account"] in am["ids_to_names"].keys():
                account_id = request["account"]

            # If it was a "friendly" name, then get the account ID for it.
            else:
                # Was this a bogus name?
                if not am["names_to_ids"].get(request["account"]):
                    stats.count(
                        "GetCredentialsHandler.post",
                        tags={
                            "user": self.user,
                            "user_role": True,
                            "account": request["account"],
                        },
                    )
                    log_data["message"] = "Can't find the passed in account."
                    log.warning(log_data)
                    error = {
                        "code": "906",
                        "message": "No matching account.",
                        "account": request["account"],
                        "request_id": self.request_uuid,
                    }
                    self.set_status(400)
                    self.write(error)
                    await self.finish()
                    return ""

                account_id = am["names_to_ids"][request["account"]]

            # Shove the account ID into the request:
            request["account_id"] = account_id

            return f"arn:aws:iam::{account_id}:role/{self.user_role_name}"

    async def post(self):
        """/api/v1/get_credentials - Endpoint used to get credentials via mtls. Used by newt and weep.
        ---
        get:
            description: Credentials endpoint. Authenticates user via MTLS and returns requested credentials.
            responses:
                200:
                    description: Returns credentials or list of matching roles
                403:
                    description: No matching roles found, or user has failed authn/authz.
        """
        log_data = {
            "function": "GetCredentialsHandler.post",
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
        }

        # Validate the input:
        data = tornado.escape.json_decode(self.request.body)
        try:
            request = await sync_to_async(credentials_schema.load)(data)
        except ValidationError as ve:
            stats.count(
                "GetCredentialsHandler.post",
                tags={"user": self.user, "validation_error": str(ve)},
            )

            log_data["validation_error"]: ve.messages
            log.error(log_data)

            error = {
                "code": "904",
                "message": f"Invalid JSON sent to the server:\n{json.dumps(ve.messages, indent=2)}",
                "request_id": self.request_uuid,
            }
            self.set_status(400)
            self.write(error)
            await self.finish()
            return
        requester_type = self.requester.get("type")

        if requester_type == "application":
            app_name = self.requester.get("name")
            await self.get_credentials_app_flow(
                app_name, self.requester, request, log_data
            )
        elif requester_type == "user":
            user_email = self.requester.get("email")
            await self.get_credentials_user_flow(user_email, request, log_data)
        else:
            raise tornado.web.HTTPError(403, "Unauthorized entity.")
        return

    async def get_credentials_app_flow(self, app_name, app, request, log_data):
        requested_role = request["requested_role"]
        log_data["requested_role"] = requested_role
        log_data["app"] = app_name
        log_data["message"] = "App is requesting role"
        log_data["custom_ip_restrictions"] = request.get("custom_ip_restrictions")
        log_data["request"] = json.dumps(request)
        log.debug(log_data)
        arn_parts = requested_role.split(":")
        if (
            len(arn_parts) != 6
            or arn_parts[0] != "arn"
            or arn_parts[1] != "aws"
            or arn_parts[2] != "iam"
        ):
            log_data["message"] = "Invalid Role ARN"
            log.warning(log_data)
            error = {
                "code": "899",
                "message": "Invalid Role ARN. Applications must pass the full role ARN when requesting credentials",
                "requested_role": requested_role,
            }
            self.set_status(403)
            self.write(error)
            await self.finish()
            return
        # Check if role is valid ARN
        authorized = await internal_config.is_context_authorized(app, requested_role)

        stats.count(
            "GetCredentialsHandler.post",
            tags={
                "user": app_name,
                "requested_role": requested_role,
                "authorized": authorized,
            },
        )

        if not authorized:
            log_data["message"] = "Unauthorized"
            log.warning(log_data)
            error = {
                "code": "900",
                "message": "Unauthorized",
                "requested_role": requested_role,
            }
            self.set_status(403)
            self.write(error)
            await self.finish()
            return

        credentials = await aws.get_credentials(
            app_name,
            requested_role,
            enforce_ip_restrictions=False,
            user_role=False,
            account_id=None,
            custom_ip_restrictions=request.get("custom_ip_restrictions"),
        )
        self.set_header("Content-Type", "application/json")
        credentials.pop("ResponseMetadata", None)
        credentials.pop("AssumedRoleUser", None)
        credentials.pop("PackedPolicySize", None)
        # Need to use ujson here because the credentials contain a datetime element
        self.write(json.dumps(credentials))
        await self.finish()
        return

    async def get_credentials_user_flow(self, user_email, request, log_data):
        log_data["user"] = user_email

        await self.authorization_flow(
            user=user_email, console_only=request["console_only"]
        )
        # Get the role to request:
        requested_role = await self._get_the_requested_role(request, log_data)
        if not requested_role:
            raise tornado.web.HTTPError(403, "No requested role detected.")
        log_data["requested_role"] = requested_role

        log_data["message"] = "User is requesting role"
        log.debug(log_data)
        matching_roles = await group_mapping.filter_eligible_roles(requested_role, self)

        log_data["matching_roles"] = matching_roles

        if len(matching_roles) == 0:
            stats.count(
                "GetCredentialsHandler.post",
                tags={"user": self.user, "requested_role": None, "authorized": False},
            )
            log_data["message"] = "No matching roles"
            log.warning(log_data)
            error = {
                "code": "900",
                "message": "No matching roles",
                "requested_role": requested_role,
                "request_id": self.request_uuid,
            }
            self.set_status(403)
            self.write(error)
            return
        if len(matching_roles) > 1:
            stats.count(
                "GetCredentialsHandler.post",
                tags={"user": self.user, "requested_role": None, "authorized": False},
            )
            log_data["message"] = "More than one matching role"
            log.warning(log_data)
            error = {
                "code": "901",
                "message": log_data["message"],
                "requested_role": requested_role,
                "matching roles": matching_roles,
                "request_id": self.request_uuid,
            }
            self.set_status(403)
            self.write(error)
            return
        if len(matching_roles) == 1:
            await self.raise_if_certificate_too_old(matching_roles[0], log_data)
            try:
                enforce_ip_restrictions = True
                if request["no_ip_restrictions"]:
                    # Duo prompt the user in order to get non IP-restricted credentials
                    mfa_success = await duo_mfa_user(
                        self.user.split("@")[0],
                        message="ConsoleMe Non-IP Restricted Credential Request",
                    )

                    if mfa_success:
                        enforce_ip_restrictions = False
                        stats.count(
                            "GetCredentialsHandler.post.no_ip_restriction.success",
                            tags={"user": self.user, "requested_role": requested_role},
                        )
                        log_data[
                            "message"
                        ] = "User requested non-IP-restricted credentials"
                        log.debug(log_data)
                    else:
                        # Log and emit a metric
                        log_data["message"] = "MFA Denied or Timeout"
                        log.warning(log_data)
                        stats.count(
                            "GetCredentialsHandler.post.no_ip_restriction.failure",
                            tags={"user": self.user, "requested_role": requested_role},
                        )
                        error = {
                            "code": "902",
                            "message": "MFA Not Successful",
                            "requested_role": requested_role,
                            "request_id": self.request_uuid,
                        }
                        self.set_status(403)
                        self.write(error)
                        await self.finish()
                        return

                log_data["enforce_ip_restrictions"] = enforce_ip_restrictions
                log_data["message"] = "Retrieving credentials"
                log.debug(log_data)

                # User-role logic:
                # User-role should come in as cm-[username or truncated username]_[N or NC]
                user_role = False
                account_id = None

                # User role must be in user's attributes
                if (
                    self.user_role_name
                    and matching_roles[0].split("role/")[1] == self.user_role_name
                ):
                    user_role = True
                    account_id = (
                        matching_roles[0].split("arn:aws:iam::")[1].split(":role")[0]
                    )

                credentials = await aws.get_credentials(
                    self.user,
                    matching_roles[0],
                    enforce_ip_restrictions=enforce_ip_restrictions,
                    user_role=user_role,
                    account_id=account_id,
                )
            except Exception as e:
                log_data["message"] = "Unable to get credentials for user"
                log_data["eligible_roles"] = self.eligible_roles
                log.error(log_data, exc_info=True)
                stats.count(
                    "GetCredentialsHandler.post.exception",
                    tags={
                        "user": self.user,
                        "requested_role": requested_role,
                        "authorized": False,
                    },
                )
                error = {
                    "code": "902",
                    "message": "Unable to get credentials.",
                    "requested_role": requested_role,
                    "matching_role": matching_roles[0],
                    "exception": str(e),
                    "request_id": self.request_uuid,
                }
                self.set_status(403)
                self.write(error)
                await self.finish()
                return
        if not credentials:
            log_data["message"] = "Unauthorized or invalid role"
            log.warning(log_data)
            stats.count(
                "GetCredentialsHandler.post.unauthorized",
                tags={
                    "user": self.user,
                    "requested_role": requested_role,
                    "authorized": False,
                },
            )
            error = {
                "code": "903",
                "message": "Requested role not found in eligible roles",
                "requested_role": requested_role,
                "eligible roles": self.eligible_roles,
                "request_id": self.request_uuid,
            }
            self.write(error)
            await self.finish()
            return
        else:
            log_data["message"] = "Success. Returning credentials"
            log.debug(log_data)
            stats.count(
                "GetCredentialsHandler.post.success",
                tags={
                    "user": self.user,
                    "requested_role": requested_role,
                    "authorized": True,
                },
            )
            credentials.pop("ResponseMetadata", None)
            credentials.pop("AssumedRoleUser", None)
            credentials.pop("PackedPolicySize", None)
            self.write(json.dumps(credentials))
            self.set_header("Content-Type", "application/json")
            await self.finish()
            return



"""
IAM stack for running ConsoleMe on ECS
"""

from aws_cdk import aws_iam as iam
from aws_cdk import aws_s3 as s3
from aws_cdk import core as cdk


class IAMStack(cdk.NestedStack):
    """
    IAM stack for running ConsoleMe on ECS
    """

    def __init__(
        self, scope: cdk.Construct, id: str, s3_bucket: s3.Bucket, **kwargs
    ) -> None:
        super().__init__(scope, id, **kwargs)

        # Define IAM roles and policies

        ecs_task_role = iam.Role(
            self,
            "TaskRole",
            role_name="ConsoleMeTaskRole",
            assumed_by=iam.ServicePrincipal("ecs-tasks.amazonaws.com"),
        )

        ecs_task_role.add_to_policy(
            iam.PolicyStatement(
                effect=iam.Effect.ALLOW,
                actions=[
                    "access-analyzer:*",
                    "cloudtrail:*",
                    "cloudwatch:*",
                    "config:SelectResourceConfig",
                    "config:SelectAggregateResourceConfig",
                    "dynamodb:batchgetitem",
                    "dynamodb:batchwriteitem",
                    "dynamodb:deleteitem",
                    "dynamodb:describe*",
                    "dynamodb:getitem",
                    "dynamodb:getrecords",
                    "dynamodb:getsharditerator",
                    "dynamodb:putitem",
                    "dynamodb:query",
                    "dynamodb:scan",
                    "dynamodb:updateitem",
                    "secretsmanager:GetResourcePolicy",
                    "secretsmanager:GetSecretValue",
                    "secretsmanager:DescribeSecret",
                    "secretsmanager:ListSecretVersionIds",
                    "secretsmanager:ListSecrets",
                    "sns:createplatformapplication",
                    "sns:createplatformendpoint",
                    "sns:deleteendpoint",
                    "sns:deleteplatformapplication",
                    "sns:getendpointattributes",
                    "sns:getplatformapplicationattributes",
                    "sns:listendpointsbyplatformapplication",
                    "sns:publish",
                    "sns:setendpointattributes",
                    "sns:setplatformapplicationattributes",
                    "sts:assumerole",
                ],
                resources=["*"],
            )
        )

        ecs_task_role.add_to_policy(
            iam.PolicyStatement(
                effect=iam.Effect.ALLOW,
                actions=["ses:sendemail", "ses:sendrawemail"],
                resources=["*"],
            )
        )

        ecs_task_role.add_to_policy(
            iam.PolicyStatement(
                effect=iam.Effect.ALLOW,
                actions=[
                    "autoscaling:Describe*",
                    "cloudwatch:Get*",
                    "cloudwatch:List*",
                    "config:BatchGet*",
                    "config:List*",
                    "config:Select*",
                    "ec2:DescribeSubnets",
                    "ec2:describevpcendpoints",
                    "ec2:DescribeVpcs",
                    "iam:GetAccountAuthorizationDetails",
                    "iam:ListAccountAliases",
                    "iam:ListAttachedRolePolicies",
                    "ec2:describeregions",
                    "s3:GetBucketPolicy",
                    "s3:GetBucketTagging",
                    "s3:ListAllMyBuckets",
                    "s3:ListBucket",
                    "s3:PutBucketPolicy",
                    "s3:PutBucketTagging",
                    "sns:GetTopicAttributes",
                    "sns:ListTagsForResource",
                    "sns:ListTopics",
                    "sns:SetTopicAttributes",
                    "sns:TagResource",
                    "sns:UnTagResource",
                    "sqs:GetQueueAttributes",
                    "sqs:GetQueueUrl",
                    "sqs:ListQueues",
                    "sqs:ListQueueTags",
                    "sqs:SetQueueAttributes",
                    "sqs:TagQueue",
                    "sqs:UntagQueue",
                ],
                resources=["*"],
            )
        )

        ecs_task_role.add_to_policy(
            iam.PolicyStatement(
                effect=iam.Effect.ALLOW,
                actions=["s3:GetObject", "s3:ListBucket"],
                resources=[s3_bucket.bucket_arn, s3_bucket.bucket_arn + "/*"],
            )
        )

        trust_role = iam.Role(
            self,
            "TrustRole",
            role_name="ConsoleMeTrustRole",
            assumed_by=iam.ArnPrincipal(arn=ecs_task_role.role_arn),
        )

        trust_role.add_to_policy(
            iam.PolicyStatement(
                effect=iam.Effect.ALLOW,
                actions=[
                    "access-analyzer:*",
                    "cloudtrail:*",
                    "cloudwatch:*",
                    "config:SelectResourceConfig",
                    "config:SelectAggregateResourceConfig",
                    "dynamodb:batchgetitem",
                    "dynamodb:batchwriteitem",
                    "dynamodb:deleteitem",
                    "dynamodb:describe*",
                    "dynamodb:getitem",
                    "dynamodb:getrecords",
                    "dynamodb:getsharditerator",
                    "dynamodb:putitem",
                    "dynamodb:query",
                    "dynamodb:scan",
                    "dynamodb:updateitem",
                    "sns:createplatformapplication",
                    "sns:createplatformendpoint",
                    "sns:deleteendpoint",
                    "sns:deleteplatformapplication",
                    "sns:getendpointattributes",
                    "sns:getplatformapplicationattributes",
                    "sns:listendpointsbyplatformapplication",
                    "sns:publish",
                    "sns:setendpointattributes",
                    "sns:setplatformapplicationattributes",
                    "sts:assumerole",
                    "autoscaling:Describe*",
                    "cloudwatch:Get*",
                    "cloudwatch:List*",
                    "config:BatchGet*",
                    "config:List*",
                    "config:Select*",
                    "ec2:DescribeSubnets",
                    "ec2:describevpcendpoints",
                    "ec2:DescribeVpcs",
                    "iam:GetAccountAuthorizationDetails",
                    "iam:ListAccountAliases",
                    "iam:ListAttachedRolePolicies",
                    "ec2:describeregions",
                    "s3:GetBucketPolicy",
                    "s3:GetBucketTagging",
                    "s3:ListAllMyBuckets",
                    "s3:ListBucket",
                    "s3:PutBucketPolicy",
                    "s3:PutBucketTagging",
                    "sns:GetTopicAttributes",
                    "sns:ListTagsForResource",
                    "sns:ListTopics",
                    "sns:SetTopicAttributes",
                    "sns:TagResource",
                    "sns:UnTagResource",
                    "sqs:GetQueueAttributes",
                    "sqs:GetQueueUrl",
                    "sqs:ListQueues",
                    "sqs:ListQueueTags",
                    "sqs:SetQueueAttributes",
                    "sqs:TagQueue",
                    "sqs:UntagQueue",
                ],
                resources=["*"],
            )
        )

        ecs_task_execution_role = iam.Role(
            self,
            "TaskExecutionRole",
            assumed_by=iam.ServicePrincipal("ecs-tasks.amazonaws.com"),
        )

        ecs_task_execution_role.add_managed_policy(
            iam.ManagedPolicy.from_managed_policy_arn(
                self,
                "ServiceRole",
                managed_policy_arn="arn:aws:iam::aws:policy/service-role/AmazonECSTaskExecutionRolePolicy",
            )
        )

        create_configuration_lambda_role = iam.Role(
            self,
            "CreateConfigurationFileLambdaRole",
            assumed_by=iam.ServicePrincipal(service="lambda.amazonaws.com"),
        )

        create_configuration_lambda_role.add_managed_policy(
            iam.ManagedPolicy.from_managed_policy_arn(
                self,
                "ConfigurationBasicExecution",
                managed_policy_arn="arn:aws:iam::aws:policy/service-role/AWSLambdaBasicExecutionRole",
            )
        )

        create_configuration_lambda_role.add_to_policy(
            iam.PolicyStatement(
                effect=iam.Effect.ALLOW,
                actions=["s3:PutObject", "s3:DeleteObject"],
                resources=[s3_bucket.bucket_arn + "/*"],
            )
        )

        self.ecs_task_role = ecs_task_role
        self.ecs_task_execution_role = ecs_task_execution_role
        self.create_configuration_lambda_role = create_configuration_lambda_role



from atlassian import Bitbucket as AtlassianBitBucketLib


class BitBucket:
    def __init__(self, url, username, password):
        self.bitbucket = AtlassianBitBucketLib(
            url=url,
            username=username,
            password=password,
        )

    async def create_pull_request(
        self,
        source_project: str,
        source_repo: str,
        dest_project: str,
        dest_repo: str,
        branch_name: str,
        main_branch_name: str,
        commit_title: str,
        commit_message: str,
    ) -> str:
        pull_request = self.bitbucket.open_pull_request(
            source_project,
            source_repo,
            dest_project,
            dest_repo,
            branch_name,
            main_branch_name,
            commit_title,
            commit_message,
        )
        pull_request_url = pull_request["links"]["self"][0]["href"]
        return pull_request_url



from unittest.mock import mock_open, patch

import ujson as json
from deepdiff import DeepDiff
from tornado.testing import AsyncHTTPTestCase


class TestRequestsHandler(AsyncHTTPTestCase):
    def get_app(self):
        from consoleme.config import config

        self.config = config

        from consoleme.routes import make_app

        return make_app(jwt_validator=lambda x: {})

    def test_get(self):
        # Method not allowed
        headers = {
            self.config.get("auth.user_header_name"): "user@github.com",
            self.config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        }
        response = self.fetch("/api/v2/requests", method="GET", headers=headers)
        self.assertEqual(response.code, 405)

    def test_requestshandler_post(self):
        mock_request_data = [
            {
                "request_id": 12345,
                "username": "user@example.com",
                "request_time": 22345,
            },
            {
                "request_id": 12346,
                "username": "userb@example.com",
                "request_time": 12345,
            },
        ]

        expected_response = {
            "totalCount": 2,
            "filteredCount": 2,
            "data": mock_request_data,
        }

        from consoleme.lib.redis import RedisHandler

        # Mocked by fakeredis
        red = RedisHandler().redis_sync()
        red.set(
            self.config.get("cache_policy_requests.redis_key", "ALL_POLICY_REQUESTS"),
            json.dumps(mock_request_data),
        )

        headers = {
            self.config.get("auth.user_header_name"): "user@github.com",
            self.config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        }
        response = self.fetch(
            "/api/v2/requests", method="POST", headers=headers, body="{}"
        )
        self.assertEqual(response.code, 200)
        diff = DeepDiff(json.loads(response.body), expected_response)
        self.assertFalse(diff)

    def test_post_request(self):
        mock_request_data = {
            "justification": "test asdf",
            "admin_auto_approve": False,
            "changes": {
                "changes": [
                    {
                        "principal": {
                            "principal_arn": "arn:aws:iam::123456789012:role/TestInstanceProfile",
                            "principal_type": "AwsResource",
                        },
                        "change_type": "inline_policy",
                        "action": "attach",
                        "policy": {
                            "policy_document": {
                                "Version": "2012-10-17",
                                "Statement": [
                                    {
                                        "Action": ["sqs:SetQueueAttributes"],
                                        "Effect": "Allow",
                                        "Resource": [
                                            "arn:aws:sqs:us-east-1:223456789012:queue"
                                        ],
                                    }
                                ],
                            }
                        },
                    }
                ]
            },
        }

        from consoleme.lib.redis import RedisHandler

        # Mocked by fakeredis
        red = RedisHandler().redis_sync()
        red.set(
            self.config.get("cache_policy_requests.redis_key", "ALL_POLICY_REQUESTS"),
            json.dumps(mock_request_data),
        )

        headers = {
            self.config.get("auth.user_header_name"): "user@github.com",
            self.config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        }
        response = self.fetch(
            "/api/v2/request",
            method="POST",
            headers=headers,
            body=json.dumps(mock_request_data),
        )
        self.assertEqual(response.code, 200)
        response_d = json.loads(response.body)
        self.assertEqual(response_d["errors"], 0)
        self.assertEqual(response_d["request_created"], True)
        self.assertIn("/policies/request/", response_d["request_url"])

    def test_post_request_admin_auto_approve(self):
        mock_request_data = {
            "justification": "test asdf",
            "admin_auto_approve": True,
            "changes": {
                "changes": [
                    {
                        "principal": {
                            "principal_arn": "arn:aws:iam::123456789012:role/TestInstanceProfile",
                            "principal_type": "AwsResource",
                        },
                        "change_type": "inline_policy",
                        "action": "attach",
                        "policy": {
                            "policy_document": {
                                "Version": "2012-10-17",
                                "Statement": [
                                    {
                                        "Action": ["sqs:SetQueueAttributes"],
                                        "Effect": "Allow",
                                        "Resource": [
                                            "arn:aws:sqs:us-east-1:223456789012:queue"
                                        ],
                                    }
                                ],
                            }
                        },
                    }
                ]
            },
        }

        from consoleme.lib.redis import RedisHandler

        # Mocked by fakeredis
        red = RedisHandler().redis_sync()
        red.set(
            self.config.get("cache_policy_requests.redis_key", "ALL_POLICY_REQUESTS"),
            json.dumps(mock_request_data),
        )

        headers = {
            self.config.get("auth.user_header_name"): "consoleme_admins@example.com",
            self.config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        }
        response = self.fetch(
            "/api/v2/request",
            method="POST",
            headers=headers,
            body=json.dumps(mock_request_data),
        )
        self.assertEqual(response.code, 200)
        response_d = json.loads(response.body)
        self.assertEqual(response_d["errors"], 0)
        self.assertEqual(response_d["request_created"], True)
        self.assertIn("/policies/request/", response_d["request_url"])
        self.assertIn(
            {"status": "success", "message": "Successfully updated request status"},
            response_d["action_results"],
        )
        self.assertIn(
            {"status": "success", "message": "Successfully updated change in dynamo"},
            response_d["action_results"],
        )

    def test_post_limit(self):
        mock_request_data = [
            {"request_id": 12345, "username": "user@example.com"},
            {"request_id": 12346, "username": "userb@example.com"},
        ]

        from consoleme.lib.redis import RedisHandler

        # Mocked by fakeredis
        red = RedisHandler().redis_sync()
        red.set(
            self.config.get("cache_policy_requests.redis_key", "ALL_POLICY_REQUESTS"),
            json.dumps(mock_request_data),
        )

        headers = {
            self.config.get("auth.user_header_name"): "user@github.com",
            self.config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        }
        response = self.fetch(
            "/api/v2/requests",
            method="POST",
            headers=headers,
            body=json.dumps({"limit": 1}),
        )
        self.assertEqual(response.code, 200)
        self.assertEqual(len(json.loads(response.body)), 3)
        self.assertEqual(len(json.loads(response.body)["data"]), 1)

    def test_post_filter(self):
        mock_request_data = [
            {"request_id": 12345, "username": "user@example.com"},
            {"request_id": 12346, "username": "userb@example.com"},
        ]

        from consoleme.lib.redis import RedisHandler

        # Mocked by fakeredis
        red = RedisHandler().redis_sync()
        red.set(
            self.config.get("cache_policy_requests.redis_key", "ALL_POLICY_REQUESTS"),
            json.dumps(mock_request_data),
        )

        headers = {
            self.config.get("auth.user_header_name"): "user@github.com",
            self.config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        }
        response = self.fetch(
            "/api/v2/requests",
            method="POST",
            headers=headers,
            body=json.dumps({"filters": {"request_id": "12346"}}),
        )
        self.assertEqual(response.code, 200)
        res = json.loads(response.body)
        self.assertEqual(len(json.loads(response.body)), 3)
        self.assertEqual(len(json.loads(response.body)["data"]), 1)
        self.assertEqual(res["data"][0], mock_request_data[1])

    def test_post_new_managed_policy_resource_request(self):
        headers = {
            self.config.get("auth.user_header_name"): "user@github.com",
            self.config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        }

        input_body = {
            "admin_auto_approve": False,
            "changes": {
                "changes": [
                    {
                        "principal": {
                            "principal_type": "AwsResource",
                            "principal_arn": "arn:aws:iam::123456789012:policy/testpolicy",
                        },
                        "change_type": "managed_policy_resource",
                        "new": True,
                        "action": "update",
                        "policy": {
                            "policy_document": {
                                "Version": "2012-10-17",
                                "Statement": [
                                    {
                                        "Action": [
                                            "s3:GetObjectVersionTagging",
                                            "s3:GetObjectAcl",
                                            "s3:ListBucket",
                                            "s3:GetObject",
                                            "s3:GetObjectVersionAcl",
                                            "s3:GetObjectTagging",
                                            "s3:GetObjectVersion",
                                            "s3:ListBucketVersions",
                                        ],
                                        "Effect": "Allow",
                                        "Resource": [
                                            "arn:aws:s3:::12345",
                                            "arn:aws:s3:::12345/*",
                                        ],
                                    }
                                ],
                            }
                        },
                    }
                ]
            },
        }
        response = self.fetch(
            "/api/v2/request",
            method="POST",
            headers=headers,
            body=json.dumps(input_body),
        )
        result = json.loads(response.body)
        result.pop("request_id")
        result.pop("request_url")
        result["extended_request"].pop("id")
        result["extended_request"].pop("timestamp")
        result["extended_request"]["changes"]["changes"][0].pop("id")
        self.assertEqual(
            result,
            {
                "errors": 0,
                "request_created": True,
                "action_results": [],
                "extended_request": {
                    "request_url": None,
                    "principal": {
                        "principal_type": "AwsResource",
                        "principal_arn": "arn:aws:iam::123456789012:policy/testpolicy",
                    },
                    "justification": None,
                    "requester_email": "user@github.com",
                    "approvers": [],
                    "request_status": "pending",
                    "cross_account": False,
                    "arn_url": "/policies/edit/123456789012/managed_policy/testpolicy",
                    "admin_auto_approve": False,
                    "changes": {
                        "changes": [
                            {
                                "principal": {
                                    "principal_type": "AwsResource",
                                    "principal_arn": "arn:aws:iam::123456789012:policy/testpolicy",
                                },
                                "change_type": "managed_policy_resource",
                                "resources": [],
                                "version": "3.0",
                                "status": "not_applied",
                                "autogenerated": False,
                                "updated_by": None,
                                "new": True,
                                "policy": {
                                    "version": None,
                                    "policy_document": {
                                        "Version": "2012-10-17",
                                        "Statement": [
                                            {
                                                "Action": [
                                                    "s3:GetObjectVersionTagging",
                                                    "s3:GetObjectAcl",
                                                    "s3:ListBucket",
                                                    "s3:GetObject",
                                                    "s3:GetObjectVersionAcl",
                                                    "s3:GetObjectTagging",
                                                    "s3:GetObjectVersion",
                                                    "s3:ListBucketVersions",
                                                ],
                                                "Effect": "Allow",
                                                "Resource": [
                                                    "arn:aws:s3:::12345",
                                                    "arn:aws:s3:::12345/*",
                                                ],
                                            }
                                        ],
                                    },
                                    "policy_sha256": None,
                                },
                                "old_policy": None,
                            }
                        ]
                    },
                    "requester_info": {
                        "email": "user@github.com",
                        "extended_info": {
                            "domain": "",
                            "userName": "user@github.com",
                            "name": {"givenName": "", "familyName": "", "fullName": ""},
                            "primaryEmail": "user@github.com",
                        },
                        "details_url": None,
                        "photo_url": "https://www.gravatar.com/avatar/1496f7f4fd086e2d0a0460220331e9ec?d=mp",
                    },
                    "reviewer": None,
                    "comments": [],
                },
            },
        )

    def test_post_new_managed_policy_resource_request_autoapprove(self):
        headers = {
            self.config.get("auth.user_header_name"): "consoleme_admins@example.com",
            self.config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        }

        input_body = {
            "admin_auto_approve": True,
            "changes": {
                "changes": [
                    {
                        "principal": {
                            "principal_type": "AwsResource",
                            "principal_arn": "arn:aws:iam::123456789012:policy/randompath/extra/testpolicy",
                        },
                        "change_type": "managed_policy_resource",
                        "new": True,
                        "action": "update",
                        "policy": {
                            "policy_document": {
                                "Version": "2012-10-17",
                                "Statement": [
                                    {
                                        "Action": [
                                            "s3:GetObjectVersionTagging",
                                            "s3:GetObjectAcl",
                                            "s3:ListBucket",
                                            "s3:GetObject",
                                            "s3:GetObjectVersionAcl",
                                            "s3:GetObjectTagging",
                                            "s3:GetObjectVersion",
                                            "s3:ListBucketVersions",
                                        ],
                                        "Effect": "Allow",
                                        "Resource": [
                                            "arn:aws:s3:::12345",
                                            "arn:aws:s3:::12345/*",
                                        ],
                                    }
                                ],
                            }
                        },
                    }
                ]
            },
        }
        response = self.fetch(
            "/api/v2/request",
            method="POST",
            headers=headers,
            body=json.dumps(input_body),
        )
        result = json.loads(response.body)
        result.pop("request_id")
        result.pop("request_url")
        result["extended_request"].pop("id")
        result["extended_request"].pop("timestamp")
        result["extended_request"]["changes"]["changes"][0].pop("id")
        result["extended_request"].pop("comments")
        self.assertEqual(
            result,
            {
                "errors": 0,
                "request_created": True,
                "action_results": [
                    {
                        "status": "success",
                        "message": "Successfully created managed policy arn:aws:iam::123456789012:policy/randompath/extra/testpolicy",
                    },
                    {
                        "status": "success",
                        "message": "Successfully updated change in dynamo",
                    },
                    {
                        "status": "success",
                        "message": "Successfully updated request status",
                    },
                ],
                "extended_request": {
                    "request_url": None,
                    "principal": {
                        "principal_type": "AwsResource",
                        "principal_arn": "arn:aws:iam::123456789012:policy/randompath/extra/testpolicy",
                    },
                    "justification": None,
                    "requester_email": "consoleme_admins@example.com",
                    "approvers": [],
                    "request_status": "approved",
                    "cross_account": False,
                    "arn_url": "/policies/edit/123456789012/managed_policy/randompath/extra/testpolicy",
                    "admin_auto_approve": True,
                    "changes": {
                        "changes": [
                            {
                                "principal": {
                                    "principal_type": "AwsResource",
                                    "principal_arn": "arn:aws:iam::123456789012:policy/randompath/extra/testpolicy",
                                },
                                "change_type": "managed_policy_resource",
                                "resources": [],
                                "version": "3.0",
                                "status": "applied",
                                "autogenerated": False,
                                "updated_by": "consoleme_admins@example.com",
                                "new": True,
                                "policy": {
                                    "version": None,
                                    "policy_document": {
                                        "Version": "2012-10-17",
                                        "Statement": [
                                            {
                                                "Action": [
                                                    "s3:GetObjectVersionTagging",
                                                    "s3:GetObjectAcl",
                                                    "s3:ListBucket",
                                                    "s3:GetObject",
                                                    "s3:GetObjectVersionAcl",
                                                    "s3:GetObjectTagging",
                                                    "s3:GetObjectVersion",
                                                    "s3:ListBucketVersions",
                                                ],
                                                "Effect": "Allow",
                                                "Resource": [
                                                    "arn:aws:s3:::12345",
                                                    "arn:aws:s3:::12345/*",
                                                ],
                                            }
                                        ],
                                    },
                                    "policy_sha256": None,
                                },
                                "old_policy": None,
                            }
                        ]
                    },
                    "requester_info": {
                        "email": "consoleme_admins@example.com",
                        "extended_info": {
                            "domain": "",
                            "userName": "consoleme_admins@example.com",
                            "name": {"givenName": "", "familyName": "", "fullName": ""},
                            "primaryEmail": "consoleme_admins@example.com",
                        },
                        "details_url": None,
                        "photo_url": "https://www.gravatar.com/avatar/ec2ee26a6397f686011678e50aeb4e81?d=mp",
                    },
                    "reviewer": "consoleme_admins@example.com",
                },
            },
        )

    def test_post_iam_role_request_dry_run(self):
        headers = {
            self.config.get("auth.user_header_name"): "user@github.com",
            self.config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        }

        input_body = {
            "dry_run": True,
            "changes": {
                "changes": [
                    {
                        "principal": {
                            "principal_type": "AwsResource",
                            "principal_arn": "arn:aws:iam::123456789012:role/RoleNumber1",
                        },
                        "change_type": "inline_policy",
                        "action": "attach",
                        "policy": {
                            "policy_document": {
                                "Version": "2012-10-17",
                                "Statement": [
                                    {
                                        "Action": ["sqs:*"],
                                        "Effect": "Allow",
                                        "Resource": ["*"],
                                    }
                                ],
                            }
                        },
                    }
                ]
            },
        }
        response = self.fetch(
            "/api/v2/request",
            method="POST",
            headers=headers,
            body=json.dumps(input_body),
        )
        result = json.loads(response.body)
        result["extended_request"].pop("id")
        result["extended_request"].pop("timestamp")
        result["extended_request"]["changes"]["changes"][0].pop("id")
        result["extended_request"]["changes"]["changes"][0].pop("policy_name")

        self.assertEqual(
            result,
            {
                "errors": 0,
                "request_created": False,
                "request_id": None,
                "request_url": None,
                "action_results": None,
                "extended_request": {
                    "request_url": None,
                    "principal": {
                        "principal_type": "AwsResource",
                        "principal_arn": "arn:aws:iam::123456789012:role/RoleNumber1",
                    },
                    "justification": None,
                    "requester_email": "user@github.com",
                    "approvers": [],
                    "request_status": "pending",
                    "cross_account": False,
                    "arn_url": "/policies/edit/123456789012/iamrole/RoleNumber1",
                    "admin_auto_approve": False,
                    "changes": {
                        "changes": [
                            {
                                "principal": {
                                    "principal_type": "AwsResource",
                                    "principal_arn": "arn:aws:iam::123456789012:role/RoleNumber1",
                                },
                                "change_type": "inline_policy",
                                "resources": [],
                                "version": "3.0",
                                "status": "not_applied",
                                "autogenerated": False,
                                "updated_by": None,
                                "new": True,
                                "action": "attach",
                                "policy": {
                                    "version": None,
                                    "policy_document": {
                                        "Version": "2012-10-17",
                                        "Statement": [
                                            {
                                                "Action": ["sqs:*"],
                                                "Effect": "Allow",
                                                "Resource": ["*"],
                                            }
                                        ],
                                    },
                                    "policy_sha256": None,
                                },
                                "old_policy": None,
                            }
                        ]
                    },
                    "requester_info": {
                        "email": "user@github.com",
                        "extended_info": {
                            "domain": "",
                            "userName": "user@github.com",
                            "name": {"givenName": "", "familyName": "", "fullName": ""},
                            "primaryEmail": "user@github.com",
                        },
                        "details_url": None,
                        "photo_url": "https://www.gravatar.com/avatar/1496f7f4fd086e2d0a0460220331e9ec?d=mp",
                    },
                    "reviewer": None,
                    "comments": [],
                },
            },
        )

    @patch("git.Repo")
    @patch("git.Git")
    def test_post_honeybee_request_dry_run(self, mock_git, mock_repo):
        headers = {
            self.config.get("auth.user_header_name"): "user@github.com",
            self.config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        }
        input_body = {
            "dry_run": True,
            "changes": {
                "changes": [
                    {
                        "principal": {
                            "principal_type": "HoneybeeAwsResourceTemplate",
                            "repository_name": "honeybee_templates",
                            "resource_identifier": "iamrole/abc.yaml",
                            "resource_url": "http://example.com/fake_repo/path/to/file.yaml",
                        },
                        "change_type": "inline_policy",
                        "action": "attach",
                        "policy": {
                            "policy_document": {
                                "Version": "2012-10-17",
                                "Statement": [
                                    {
                                        "Action": [
                                            "s3:GetObjectVersionTagging",
                                            "s3:GetObjectAcl",
                                            "s3:ListBucket",
                                            "s3:GetObject",
                                            "s3:GetObjectVersionAcl",
                                            "s3:GetObjectTagging",
                                            "s3:GetObjectVersion",
                                            "s3:ListBucketVersions",
                                        ],
                                        "Effect": "Allow",
                                        "Resource": [
                                            "arn:aws:s3:::bucketa",
                                            "arn:aws:s3:::bucketa/*",
                                        ],
                                        "Sid": "cmccastrapel1623864565xhwu",
                                    }
                                ],
                            }
                        },
                    }
                ]
            },
        }
        template_data = """
Policies:
  - IncludeAccounts:
      - account_a
      - account_b
      - account_c
    PolicyName: policy_a
    Statement:
      - Action:
          - '*'
        Effect: Allow
        Resource:
          - '*'
        Sid: admin"""
        with patch("builtins.open", mock_open(read_data=template_data)):
            response = self.fetch(
                "/api/v2/request",
                method="POST",
                headers=headers,
                body=json.dumps(input_body),
            )
            result = json.loads(response.body)
            result["extended_request"].pop("timestamp")
            result["extended_request"].pop("id")
            yaml_policy = result["extended_request"]["changes"]["changes"][0].pop(
                "policy"
            )
            from consoleme.lib.yaml import yaml

            # Get this in a standard dictionary format
            generated_policy = json.loads(json.dumps(yaml.load(yaml_policy)))
            generated_policy["Policies"][1]["Statement"][0].pop("Sid")
            self.assertEqual(
                generated_policy,
                {
                    "Policies": [
                        {
                            "IncludeAccounts": ["account_a", "account_b", "account_c"],
                            "PolicyName": "policy_a",
                            "Statement": [
                                {
                                    "Action": ["*"],
                                    "Effect": "Allow",
                                    "Resource": ["*"],
                                    "Sid": "admin",
                                }
                            ],
                        },
                        {
                            "PolicyName": "self_service_generated",
                            "Statement": [
                                {
                                    "Action": [
                                        "s3:GetObjectVersionTagging",
                                        "s3:GetObjectAcl",
                                        "s3:ListBucket",
                                        "s3:GetObject",
                                        "s3:GetObjectVersionAcl",
                                        "s3:GetObjectTagging",
                                        "s3:GetObjectVersion",
                                        "s3:ListBucketVersions",
                                    ],
                                    "Effect": "Allow",
                                    "Resource": [
                                        "arn:aws:s3:::bucketa",
                                        "arn:aws:s3:::bucketa/*",
                                    ],
                                }
                            ],
                        },
                    ]
                },
            )
            self.assertEqual(
                result,
                {
                    "errors": 0,
                    "request_created": False,
                    "request_id": None,
                    "request_url": None,
                    "action_results": None,
                    "extended_request": {
                        "request_url": "",
                        "principal": {
                            "principal_type": "HoneybeeAwsResourceTemplate",
                            "repository_name": "honeybee_templates",
                            "resource_identifier": "iamrole/abc.yaml",
                            "resource_url": "http://example.com/fake_repo/path/to/file.yaml",
                        },
                        "justification": None,
                        "requester_email": "user@github.com",
                        "approvers": [],
                        "request_status": "pending",
                        "cross_account": False,
                        "arn_url": None,
                        "admin_auto_approve": False,
                        "changes": {
                            "changes": [
                                {
                                    "principal": {
                                        "principal_type": "HoneybeeAwsResourceTemplate",
                                        "repository_name": "honeybee_templates",
                                        "resource_identifier": "iamrole/abc.yaml",
                                        "resource_url": "http://example.com/fake_repo/path/to/file.yaml",
                                    },
                                    "change_type": "generic_file",
                                    "resources": [],
                                    "version": 3.0,
                                    "status": "not_applied",
                                    "id": None,
                                    "autogenerated": False,
                                    "updated_by": None,
                                    "action": "attach",
                                    "old_policy": "Policies:\n  - IncludeAccounts:\n      - account_a\n      - account_b\n      - account_c\n    PolicyName: policy_a\n    Statement:\n      - Action:\n          - '*'\n        Effect: Allow\n        Resource:\n          - '*'\n        Sid: admin\n",
                                    "encoding": "yaml",
                                }
                            ]
                        },
                        "requester_info": {
                            "email": "user@github.com",
                            "extended_info": {
                                "domain": "",
                                "userName": "user@github.com",
                                "name": {
                                    "givenName": "",
                                    "familyName": "",
                                    "fullName": "",
                                },
                                "primaryEmail": "user@github.com",
                            },
                            "details_url": None,
                            "photo_url": "https://www.gravatar.com/avatar/1496f7f4fd086e2d0a0460220331e9ec?d=mp",
                        },
                        "reviewer": None,
                        "comments": [],
                    },
                },
            )


class TestRequestDetailHandler(AsyncHTTPTestCase):
    def get_app(self):
        from consoleme.routes import make_app

        return make_app(jwt_validator=lambda x: {})

    def test_get(self):
        # expected = {
        #     "status": 501,
        #     "title": "Not Implemented",
        #     "message": "Get request details",
        # }
        # headers = {
        #     config.get("auth.user_header_name"): "user@github.com",
        #     config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        # }
        # response = self.fetch(
        #     "/api/v2/requests/16fd2706-8baf-433b-82eb-8c7fada847da",
        #     method="GET",
        #     headers=headers,
        # )
        # TODO: add unit tests
        pass
        # self.assertEqual(response.code, 501)
        # self.assertDictEqual(json.loads(response.body), expected)

    def test_put(self):
        # expected = {
        #     "status": 501,
        #     "title": "Not Implemented",
        #     "message": "Update request details",
        # }
        # headers = {
        #     config.get("auth.user_header_name"): "user@github.com",
        #     config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        # }
        # response = self.fetch(
        #     "/api/v2/requests/16fd2706-8baf-433b-82eb-8c7fada847da",
        #     method="PUT",
        #     headers=headers,
        #     body="{}",
        # )
        # self.assertEqual(response.code, 501)
        # self.assertDictEqual(json.loads(response.body), expected)
        # TODO: add unit tests
        pass



import sys

from consoleme.config import config
from consoleme.handlers.base import BaseMtlsHandler
from consoleme.lib.cloud_credential_authorization_mapping import (
    CredentialAuthorizationMapping,
)
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.models import Status2, WebResponse

log = config.get_logger()
stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()
credential_mapping = CredentialAuthorizationMapping()


def _get_last_page(total: int, page_size: int) -> int:
    pages = int(total / page_size)
    if not total % page_size:
        pages += 1
    return pages


class AuditRolesHandler(BaseMtlsHandler):
    """Handler for /api/v2/audit/roles

    Returns a list of all roles known to ConsoleMe
    """

    allowed_methods = ["GET"]

    def check_xsrf_cookie(self) -> None:
        pass

    async def get(self):
        """
        GET /api/v2/audit/roles
        """
        log_data = {
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
        }

        page = self.get_argument("page", "0")
        try:
            page = int(page)
        except ValueError:
            log_data["message"] = f"invalid value for page: {page}"
            log.warning(log_data)
            page = 0

        count = self.get_argument("count", "1000")
        try:
            count = int(count)
        except ValueError:
            log_data["message"] = f"invalid value for count: {count}"
            log.warning(log_data)
            count = 1000

        if page < 0:
            page = 0
        if count <= 0:
            count = 1000

        app_name = self.requester.get("name") or self.requester.get("username")
        stats.count(
            "AuditRoleHandler.get",
            tags={
                "requester": app_name,
            },
        )

        roles = await credential_mapping.all_roles(
            paginate=True, page=page, count=count
        )
        total_roles = await credential_mapping.number_roles()
        start = page * count
        end = start + count
        end = min(end, total_roles)
        roles = roles[start:end]

        self.write(
            WebResponse(
                status=Status2.success,
                status_code=200,
                data=roles,
                page=page,
                total=total_roles,
                count=len(roles),
                last_page=_get_last_page(total_roles, count),
            ).json(exclude_unset=True)
        )


class AuditRolesAccessHandler(BaseMtlsHandler):
    """Handler for /api/v2/audit/roles/{accountNumber}/{roleName}/access

    Returns a list of groups with access to the requested role
    """

    allowed_methods = ["GET"]

    def check_xsrf_cookie(self) -> None:
        pass

    async def get(self, account_id, role_name):
        """
        GET /api/v2/audit/roles/{accountNumber}/{roleName}/access
        """
        log_data = {
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
            "account_id": account_id,
            "role_name": role_name,
        }
        app_name = self.requester.get("name") or self.requester.get("username")
        stats.count(
            "RoleAccessHandler.get",
            tags={
                "requester": app_name,
                "account_id": account_id,
                "role_name": role_name,
            },
        )

        groups = await credential_mapping.determine_role_authorized_groups(
            account_id, role_name
        )
        if not groups:
            log_data[
                "message"
            ] = f"No authorized groups found for {role_name} in {account_id}"
            log.warning(log_data)
            self.set_status(404)
            self.write(
                WebResponse(
                    status=Status2.error,
                    status_code=404,
                    message="No groups found for requested role",
                ).json(exclude_unset=True)
            )
            return

        self.write(
            WebResponse(status=Status2.success, status_code=200, data=groups).json(
                exclude_unset=True
            )
        )



"""Web routes."""
import os

import pkg_resources
import requests
import sentry_sdk
import tornado.autoreload
import tornado.web
from sentry_sdk.integrations.aiohttp import AioHttpIntegration
from sentry_sdk.integrations.redis import RedisIntegration
from sentry_sdk.integrations.tornado import TornadoIntegration

import consoleme
from consoleme.config import config
from consoleme.handlers.auth import AuthHandler
from consoleme.handlers.v1.credentials import GetCredentialsHandler
from consoleme.handlers.v1.headers import ApiHeaderHandler, HeaderHandler
from consoleme.handlers.v1.health import HealthHandler
from consoleme.handlers.v1.policies import (
    ApiResourceTypeAheadHandler,
    AutocompleteHandler,
    ResourceTypeAheadHandler,
)
from consoleme.handlers.v1.roles import GetRolesHandler
from consoleme.handlers.v1.saml import SamlHandler
from consoleme.handlers.v2.audit import AuditRolesAccessHandler, AuditRolesHandler
from consoleme.handlers.v2.aws_iam_users import UserDetailHandler
from consoleme.handlers.v2.challenge import (
    ChallengeGeneratorHandler,
    ChallengePollerHandler,
    ChallengeValidatorHandler,
)
from consoleme.handlers.v2.dynamic_config import DynamicConfigApiHandler
from consoleme.handlers.v2.errors import NotFoundHandler as V2NotFoundHandler
from consoleme.handlers.v2.generate_changes import GenerateChangesHandler
from consoleme.handlers.v2.generate_policy import GeneratePolicyHandler
from consoleme.handlers.v2.index import (
    EligibleRoleHandler,
    EligibleRolePageConfigHandler,
    FrontendHandler,
)
from consoleme.handlers.v2.logout import LogOutHandler
from consoleme.handlers.v2.managed_policies import (
    ManagedPoliciesForAccountHandler,
    ManagedPoliciesHandler,
    ManagedPoliciesOnPrincipalHandler,
)
from consoleme.handlers.v2.notifications import NotificationsHandler
from consoleme.handlers.v2.policies import (
    CheckPoliciesHandler,
    PoliciesHandler,
    PoliciesPageConfigHandler,
)
from consoleme.handlers.v2.requests import (
    RequestDetailHandler,
    RequestHandler,
    RequestsHandler,
    RequestsPageConfigHandler,
)
from consoleme.handlers.v2.resources import GetResourceURLHandler, ResourceDetailHandler
from consoleme.handlers.v2.roles import (
    AccountRolesHandler,
    GetRolesMTLSHandler,
    RoleCloneHandler,
    RoleConsoleLoginHandler,
    RoleDetailAppHandler,
    RoleDetailHandler,
    RolesHandler,
)
from consoleme.handlers.v2.self_service import (
    PermissionTemplatesHandler,
    SelfServiceConfigHandler,
)
from consoleme.handlers.v2.service_control_policy import ServiceControlPolicyHandler
from consoleme.handlers.v2.templated_resources import TemplatedResourceDetailHandler
from consoleme.handlers.v2.typeahead import (
    ResourceTypeAheadHandlerV2,
    SelfServiceStep1ResourceTypeahead,
)
from consoleme.handlers.v2.user import (
    LoginConfigurationHandler,
    LoginHandler,
    UserManagementHandler,
    UserRegistrationHandler,
)
from consoleme.handlers.v2.user_profile import UserProfileHandler
from consoleme.lib.auth import mk_jwks_validator
from consoleme.lib.plugins import get_plugin_by_name

internal_routes = get_plugin_by_name(
    config.get("plugins.internal_routes", "default_internal_routes")
)()

log = config.get_logger()


def make_jwt_validator():
    jwk_url = config.get("sso.jwk_url")
    if not jwk_url:
        raise Exception("Config 'sso.jwk_url' is not defined")
    jwk_set = requests.get(jwk_url).json()
    keys = [k for k in jwk_set["keys"] if k["kty"] == "RSA"]
    jwk_schema = config.get("sso.jwk_schema")
    if not jwk_schema:
        raise Exception("Config 'sso.jwk_schema' is not defined")
    return mk_jwks_validator(keys, jwk_schema["header"], jwk_schema["payload"])


def make_app(jwt_validator=None):
    """make_app."""
    path = pkg_resources.resource_filename("consoleme", "templates")

    oss_routes = [
        (r"/auth", AuthHandler),
        (r"/healthcheck", HealthHandler),
        (
            r"/static/(.*)",
            tornado.web.StaticFileHandler,
            dict(path=os.path.join(path, "static")),
        ),
        (
            r"/images/(.*)",
            tornado.web.StaticFileHandler,
            dict(path=os.path.join(path, "images")),
        ),
        (
            r"/(favicon.ico)",
            tornado.web.StaticFileHandler,
            dict(path=path),
        ),
        (r"/api/v1/get_credentials", GetCredentialsHandler),
        (r"/api/v1/get_roles", GetRolesHandler),
        (r"/api/v2/get_roles", GetRolesMTLSHandler),
        (r"/api/v2/get_resource_url", GetResourceURLHandler),
        # Used to autocomplete AWS permissions
        (r"/api/v1/policyuniverse/autocomplete/?", AutocompleteHandler),
        (r"/api/v2/user_profile/?", UserProfileHandler),
        (r"/api/v2/self_service_config/?", SelfServiceConfigHandler),
        (r"/api/v2/permission_templates/?", PermissionTemplatesHandler),
        (r"/api/v1/myheaders/?", ApiHeaderHandler),
        (r"/api/v1/policies/typeahead", ApiResourceTypeAheadHandler),
        (r"/api/v2/policies/check", CheckPoliciesHandler),
        (r"/api/v2/dynamic_config", DynamicConfigApiHandler),
        (r"/api/v2/eligible_roles", EligibleRoleHandler),
        (r"/api/v2/eligible_roles_page_config", EligibleRolePageConfigHandler),
        (r"/api/v2/policies_page_config", PoliciesPageConfigHandler),
        (r"/api/v2/requests_page_config", RequestsPageConfigHandler),
        (r"/api/v2/generate_policy", GeneratePolicyHandler),
        (r"/api/v2/notifications/?", NotificationsHandler),
        (r"/api/v2/managed_policies/(\d{12})", ManagedPoliciesForAccountHandler),
        (r"/api/v2/managed_policies/(.*)", ManagedPoliciesHandler),
        (
            r"/api/v2/templated_resource/([a-zA-Z0-9_-]+)/(.*)",
            TemplatedResourceDetailHandler,
        ),
        (
            r"/api/v2/managed_policies_on_principal/(.*)",
            ManagedPoliciesOnPrincipalHandler,
        ),
        (r"/api/v2/login", LoginHandler),
        (r"/api/v2/login_configuration", LoginConfigurationHandler),
        (r"/api/v2/logout", LogOutHandler),
        (
            r"/api/v2/typeahead/self_service_resources",
            SelfServiceStep1ResourceTypeahead,
        ),
        (r"/api/v2/user", UserManagementHandler),
        (r"/api/v2/user_registration", UserRegistrationHandler),
        (r"/api/v2/policies", PoliciesHandler),
        (r"/api/v2/request", RequestHandler),
        (r"/api/v2/requests", RequestsHandler),
        (r"/api/v2/requests/([a-zA-Z0-9_-]+)", RequestDetailHandler),
        (r"/api/v2/roles/?", RolesHandler),
        (r"/api/v2/roles/(\d{12})", AccountRolesHandler),
        (r"/api/v2/roles/(\d{12})/(.*)", RoleDetailHandler),
        (r"/api/v2/users/(\d{12})/(.*)", UserDetailHandler),
        (
            r"/api/v2/resources/(\d{12})/(s3|sqs|sns|managed_policy)(?:/([a-z\-1-9]+))?/(.*)",
            ResourceDetailHandler,
        ),
        (r"/api/v2/service_control_policies/(.*)", ServiceControlPolicyHandler),
        (r"/api/v2/mtls/roles/(\d{12})/(.*)", RoleDetailAppHandler),
        (r"/api/v2/clone/role", RoleCloneHandler),
        (r"/api/v2/generate_changes/?", GenerateChangesHandler),
        (r"/api/v2/typeahead/resources", ResourceTypeAheadHandlerV2),
        (r"/api/v2/role_login/(.*)", RoleConsoleLoginHandler),
        (r"/myheaders/?", HeaderHandler),
        (r"/policies/typeahead/?", ResourceTypeAheadHandler),
        (r"/saml/(.*)", SamlHandler),
        (
            r"/api/v2/challenge_validator/([a-zA-Z0-9_-]+)",
            ChallengeValidatorHandler,
        ),
        (r"/noauth/v1/challenge_generator/(.*)", ChallengeGeneratorHandler),
        (r"/noauth/v1/challenge_poller/([a-zA-Z0-9_-]+)", ChallengePollerHandler),
        (r"/api/v2/audit/roles", AuditRolesHandler),
        (r"/api/v2/audit/roles/(\d{12})/(.*)/access", AuditRolesAccessHandler),
        (r"/api/v2/.*", V2NotFoundHandler),
        (
            r"/(.*)",
            FrontendHandler,
            dict(path=path, default_filename="index.html"),
        ),
    ]

    # Prioritize internal routes before OSS routes so that OSS routes can be overridden if desired.
    internal_route_list = internal_routes.get_internal_routes(
        make_jwt_validator, jwt_validator
    )
    routes = internal_route_list + oss_routes

    app = tornado.web.Application(
        routes,
        debug=config.get("tornado.debug", False),
        xsrf_cookies=config.get("tornado.xsrf", True),
        xsrf_cookie_kwargs=config.get("tornado.xsrf_cookie_kwargs", {}),
        template_path=config.get(
            "tornado.template_path", f"{os.path.dirname(consoleme.__file__)}/templates"
        ),
        ui_modules=internal_routes.ui_modules,
    )
    sentry_dsn = config.get("sentry.dsn")

    if sentry_dsn:
        sentry_sdk.init(
            dsn=sentry_dsn,
            integrations=[
                TornadoIntegration(),
                AioHttpIntegration(),
                RedisIntegration(),
            ],
        )

    return app



"""Entrypoint for ConsoleMe. To run service, set CONFIG_LOCATION environmental variable and run
python -m consoleme.__main__"""

import asyncio
import logging
import os

import tornado.autoreload
import tornado.httpserver
import tornado.ioloop
import uvloop
from tornado.platform.asyncio import AsyncIOMainLoop

from consoleme.config import config
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.routes import make_app

logging.basicConfig(level=logging.DEBUG, format=config.get("logging.format"))
logging.getLogger("urllib3.connectionpool").setLevel(logging.CRITICAL)
stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()
log = config.get_logger()


def main():
    if config.get("sso.create_mock_jwk"):
        app = make_app(jwt_validator=lambda x: {})
    else:
        app = make_app()
    return app


if config.get("tornado.uvloop", True):
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
AsyncIOMainLoop().install()
app = main()


def init():
    port = config.get("tornado.port")
    stats.count("start")

    server = tornado.httpserver.HTTPServer(app)

    if port:
        server.bind(port, address=config.get("tornado.address"))

    server.start()  # forks one process per cpu

    if config.get("tornado.debug", False):
        for directory, _, files in os.walk("consoleme/templates"):
            [
                tornado.autoreload.watch(directory + "/" + f)
                for f in files
                if not f.startswith(".")
            ]
    log.debug({"message": "Server started"})
    asyncio.get_event_loop().run_forever()


if __name__ == "__main__":
    init()



from consoleme.handlers.base import BaseAPIV2Handler
from consoleme.lib.templated_resources import retrieve_cached_resource_templates


class TemplatedResourceDetailHandler(BaseAPIV2Handler):
    async def get(self, repository_name, resource):
        matching_template = await retrieve_cached_resource_templates(
            repository_name=repository_name,
            resource=resource,
            return_first_result=True,
        )
        if not matching_template:
            # TODO: Log here
            # Return 404
            self.write({})
            return
        self.write(matching_template.json())
